"""Setup directory from command line with
`python -m aoc_lube [-y YEAR] [-t TEMPLATE]`.
"""
import argparse
from pathlib import Path

from . import setup_dir

parser = argparse.ArgumentParser(
    prog="python -m aoc_lube", description="Setup a directory for Advent of Code."
)
parser.add_argument("-y", "--year", help="Advent of Code year.", default=None)
parser.add_argument("-t", "--template", help="A code template file.", default=None)
args = parser.parse_args()

setup_dir(args.year, args.template if args.template is None else Path(args.template))



"""Helpful functions for AoC.

Imports are deferred so that this file loads faster. Requires `networkx` and `numpy`.
"""

from collections.abc import Iterable, Iterator, ValuesView
from typing import Final, Literal, NamedTuple, Self

__all__ = [
    "GRID_NEIGHBORHOODS",
    "UnionFind",
    "Vec2",
    "chinese_remainder_theorem",
    "chunk",
    "distribute",
    "dot_print",
    "grid_steps",
    "extract_ints",
    "extract_maze",
    "ilen",
    "int_grid",
    "maximum_matching",
    "ndigits",
    "nth",
    "oscillate_range",
    "pairwise",
    "partitions",
    "shift_cipher",
    "shiftmod",
    "sliding_window",
    "sliding_window_cycle",
    "spiral_grid",
    "split",
]


class UnionFind[T]:
    """A collection for fast unions of disjoint sets."""

    def __init__(self, iterable=Iterable[T] | None):
        self._parents: dict[T, T] = {}
        self._ranks: dict[T, int] = {}
        self._components: dict[T, set[T]] = {}

        if iterable is not None:
            for item in iterable:
                self.add(item)

    @property
    def components(self) -> ValuesView[set[T]]:
        """Return the disjoint sets."""
        return self._components.values()

    @property
    def size(self) -> int:
        """Number of items in all disjoint sets."""
        return len(self._parents)

    def __contains__(self, item: T) -> bool:
        return item in self._parents

    def __len__(self) -> int:
        return len(self._components)

    def __getitem__(self, item: T) -> T:
        if item not in self:
            raise KeyError(item)

        root = self._find(item)
        return self._components[root]

    def __iter__(self) -> Iterator[T]:
        """Yield each disjoint set."""
        yield from self._components

    def elements(self) -> Iterator[T]:
        """Yield each element of each disjoint set."""
        yield from self._parents.keys()

    def add(self, item: T) -> None:
        """Add a new item to the disjoint set forest."""
        if item in self:
            return
        self._parents[item] = item
        self._ranks[item] = 0
        self._components[item] = {item}

    def _find(self, item: T) -> T:
        """Find the representive of the item's disjoint set."""
        if item not in self:
            raise KeyError(item)

        if self._parents[item] != item:
            self._parents[item] = self._find(self._parents[item])
        return self._parents[item]

    def merge(self, a: T, b: T) -> None:
        """Merge the set containing ``a`` and the set containing ``b``."""
        a_root = self._find(a)
        b_root = self._find(b)
        if a_root == b_root:
            return

        if self._ranks[a_root] < self._ranks[b_root]:
            self._parents[a_root] = b_root
            self._components[b_root] |= self._components.pop(a_root)
        elif self._ranks[a_root] > self._ranks[b_root]:
            self._parents[b_root] = a_root
            self._components[a_root] |= self._components.pop(b_root)
        else:
            self._parents[a_root] = b_root
            self._ranks[b_root] += 1
            self._components[b_root] |= self._components.pop(a_root)


class Vec2(NamedTuple):
    """A 2D point."""

    y: int
    x: int

    def __add__(self, other: tuple[int, int]) -> Self:
        y1, x1 = self
        y2, x2 = other
        return Vec2(y1 + y2, x1 + x2)

    def __sub__(self, other: tuple[int, int]) -> Self:
        y1, x1 = self
        y2, x2 = other
        return Vec2(y1 - y2, x1 - x2)

    def __neg__(self) -> Self:
        y, x = self
        return Vec2(-y, -x)

    def __mul__(self, n: int) -> Self:
        y, x = self
        return Vec2(n * y, n * x)

    def rotate(self, clockwise: bool = True) -> Self:
        """Rotate vector 90 degrees."""
        y, x = self
        if clockwise:
            return Vec2(x, -y)
        return Vec2(-x, y)

    def inbounds(self, size: tuple[int, int], pos: tuple[int, int] = (0, 0)) -> bool:
        """Return whether vec is within some rect."""
        y, x = self
        h, w = size
        oy, ox = pos
        return 0 <= y - oy < h and 0 <= x - ox < w

    @classmethod
    def iter_rect(
        self, size: tuple[int, int], pos: tuple[int, int] = (0, 0)
    ) -> Iterator[Self]:
        """Generate all points in some rect."""
        h, w = size
        oy, ox = pos
        for y in range(h):
            for x in range(w):
                yield Vec2(y + oy, x + ox)

    def adj(self, neighborhood: Literal[4, 5, 8, 9] = 4) -> Iterator[Self]:
        """Yield adjacent points given by neighborhood."""
        return (pos + self for pos in GRID_NEIGHBORHOODS[neighborhood])


# fmt: off
GRID_NEIGHBORHOODS: Final = {
    4: [Vec2(0, 1), Vec2(0, -1), Vec2(1, 0), Vec2(-1, 0)],
    5: [Vec2(0, 0), Vec2(0, 1), Vec2(0, -1), Vec2(1, 0), Vec2(-1, 0)],
    8: [
        Vec2(0, 1), Vec2(0, -1), Vec2(1, 0), Vec2(-1, 0),
        Vec2(1, 1), Vec2(-1, -1), Vec2(1, -1), Vec2(-1, 1),
    ],
    9: [
        Vec2(0, 0), Vec2(0, 1), Vec2(0, -1), Vec2(1, 0), Vec2(-1, 0),
        Vec2(1, 1), Vec2(-1, -1), Vec2(1, -1), Vec2(-1, 1),
    ],
}
# fmt: on


def chinese_remainder_theorem(moduli, residues):
    """Find the solution to a system of modular equations."""
    from math import prod

    N = prod(moduli)

    return (
        sum(
            (div := (N // modulus)) * pow(div, -1, modulus) * residue
            for modulus, residue in zip(moduli, residues)
        )
        % N
    )


def chunk(it, n: int, fillvalue=None):
    """Chunk an iterable into non-overlapping fixed sized pieces."""
    from itertools import zip_longest

    args = [iter(it)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def distribute(iterable, n):
    """Distribute an iterable amoung `n` smaller iterables."""
    from itertools import islice, tee

    children = tee(iterable, n)
    return [islice(it, index, None, n) for index, it in enumerate(children)]


def dot_print(array):
    """Pretty print a binary or boolean array."""
    for row in array:
        print("".join(" #"[i] for i in row))


def grid_steps(neighborhood, height, width):
    """Yield all steps ((y, x), (y + dy, x + dx)) with dy, dx given by `neighborhood`.

    `neighborhood` is one of 4, 5, 8 or 9 (For Von-Neumann or Moore neighborhoods with
    or without center cell. The number is the number of cells in the neighborhood.)
    """
    deltas = GRID_NEIGHBORHOODS[neighborhood]

    for y in range(height):
        for x in range(width):
            for dy, dx in deltas:
                new_y = y + dy
                new_x = x + dx

                if 0 <= new_y < height and 0 <= new_x < width:
                    yield (y, x), (new_y, new_x)


def extract_ints(raw: str):
    """Extract integers from a string."""
    import re

    return map(int, re.findall(r"(-?\d+)", raw))


def extract_maze(raw: str, wall="#", largest_component=False):
    """Parse an ascii maze into a networkx graph. Return a tuple
    `(np.array, nx.Graph)`.
    """
    import networkx as nx
    import numpy as np

    lines = raw.splitlines()
    max_width = max(map(len, lines))
    maze = np.array([list(line + " " * (max_width - len(line))) for line in lines])

    G = nx.grid_graph(maze.shape[::-1])

    walls = np.stack(np.where(maze == wall)).T
    G.remove_nodes_from(map(tuple, walls))

    if largest_component:
        G.remove_nodes_from(
            G.nodes - max(nx.connected_components(G), key=lambda g: len(g))
        )

    return maze, G


def ilen(iterable):
    """Return number of items in `iterable`.

    This will consume the iterable.
    """
    return sum(1 for _ in iterable)


def int_grid(raw, np=True, separator=""):
    """Parse a grid of ints into a 2d list or numpy array (if np==True)."""
    array = [
        [int(i) for i in (line.split(separator) if separator else line) if i]
        for line in raw.splitlines()
    ]

    if np:
        import numpy as np

        return np.array(array)

    return array


def maximum_matching(items: dict[list]):
    """Return a maximum matching from a dict of lists."""
    import networkx as nx

    G = nx.from_dict_of_lists(items)

    for k, v in nx.bipartite.maximum_matching(G, top_nodes=items).items():
        if k in items:  # Filter edges pointing the wrong direction.
            yield k, v


def ndigits(n: int) -> int:
    """Return the number of digits in ``n``."""
    from math import log10

    return int(log10(n)) + 1


def nth(iterable, n):
    """Return nth item of `iterable`."""
    from itertools import islice

    return next(islice(iterable, n, None))


def oscillate_range(start=None, stop=None, step=None, /):
    """Yield values around start."""
    match start, stop, step:
        case (int(), None, None):
            start, stop, step = 0, start, 1 if start > 0 else -1
        case (int(), int(), None):
            step = 1 if start < stop else -1
        case (int(), int(), int()) if step != 0:
            pass
        case _:
            ValueError(f"non-integer values or 0 step ({start=}, {stop=}, {step=})")

    stop_n = (stop - start) // step

    if stop_n <= 0:
        return

    yield start

    n = 1
    while n < stop_n:
        yield start + step * n
        yield start - step * n
        n += 1


def pairwise(iterable, offset=1):
    """Return successive pairs from an iterable separated by `offset`."""
    from itertools import islice, tee

    a, b = tee(iterable)

    return zip(a, islice(b, offset, None))


def partitions(n, r):
    """Generate integer partitions of  `n` into `r` parts."""
    if r == 1:
        yield (n,)
        return

    for i in range(n + 1):
        for j in partitions(n - i, r - 1):
            yield i, *j


def shift_cipher(text, n):
    """Shift all letters `n` characters in text."""

    def _shift_letter(letter):
        if letter.isupper():
            first_ord = ord("A")
        elif letter.islower():
            first_ord = ord("a")
        else:
            return letter

        return chr(shiftmod(ord(letter) + n, 26, shift=first_ord))

    return "".join(map(_shift_letter, text))


def shiftmod(n, m, shift=1):
    """Simlar to n % m except the result lies within [shift, m + shift).

    Examples
    --------
    ```py
    shiftmod(10, 10, shift=1) == 10
    shiftmod(11, 10, shift=1) == 1
    shiftmod(11, 10, shift=2) == 11
    shiftmod(12, 10, shift=2) == 2
    ```

    """
    return (n - shift) % m + shift


def sliding_window(iterable, length=2):
    """Return a sliding window over an iterable."""
    from itertools import islice, tee

    its = (islice(it, i, None) for i, it in enumerate(tee(iterable, length)))

    return zip(*its)


def sliding_window_cycle(iterable, length=2):
    """Return a sliding window over an iterable that wraps around."""
    from itertools import chain, islice

    it = iter(iterable)
    start = tuple(islice(it, length - 1))
    return sliding_window(chain(start, it, start), length)


def spiral_grid():
    """Yield 2D coordinates spiraling around the origin."""
    x = y = 0
    d = m = 1
    while True:
        while 2 * x * d < m:
            yield x, y
            x = x + d

        while 2 * y * d < m:
            yield x, y
            y = y + d

        d *= -1
        m += 1


def split(sequence, n=2):
    """Split a sequence into `n` equal parts. `n` is assumed to divide the
    length of the sequence.
    """
    div = len(sequence) // n
    return (sequence[i * div : (i + 1) * div] for i in range(n))



"""Propert Advent of Code lubricant."""

import re
import subprocess
import time
import webbrowser
from datetime import date, datetime, timedelta, timezone
from pathlib import Path
from typing import Callable, Literal

import bs4
import requests
import tomlkit

from . import utils

__all__ = ["fetch", "setup_dir", "submit", "utils"]

__version__ = "1.3.0"

CONFIG_DIR = Path.home() / ".aoc_lube"
if not CONFIG_DIR.exists():
    CONFIG_DIR.mkdir()

HEADERS = {
    "User-Agent": (
        f"github.com/salt-die/aoc_lube v{__version__} by salt-die@protonmail.com"
    ),
}
SUBMISSIONS_FILE = "submissions.toml"
TEMPLATE_FILE = Path(__file__).parent / "code_template.txt"
TOKEN_FILE = CONFIG_DIR / ".token"
# AoC puzzle inputs unlock at midnight -5 UTC during month of December.
UNLOCK_TIME_INFO = dict(
    month=12,
    hour=0,
    minute=0,
    second=5,
    microsecond=0,
    tzinfo=timezone(timedelta(hours=-5), "Eastern US"),
)
URL = "https://adventofcode.com/{year}/day/{day}"

# Ansi Escapes
RED = "\x1b[31m"
GREEN = "\x1b[32m"
YELLOW = "\x1b[33m"
BLUE = "\x1b[34m"
BOLD = "\x1b[1m"
RESET = "\x1b[0m"
HIDE_CURSOR = "\x1b[?25l"
SHOW_CURSOR = "\x1b[?25h"

try:
    TOKEN = {"session": TOKEN_FILE.read_text().strip()}
except FileNotFoundError:
    print(
        f"\n{BOLD}{RED}::WARNING::{RESET}\n\n"
        f"Token not found at {BLUE}{TOKEN_FILE.absolute()}{RESET}.\n"
        "`fetch` and `submit` functions will fail without a user token.\n"
        f"See {YELLOW}README{RESET} for instructions on how to get your user token.\n"
    )


def setup_dir(year: int | None = None, template: Path | None = None) -> None:
    r"""Run once to setup directory with templates for daily solutions.

    A file can be provided as a template. The file's text should be able to be formatted
    with `year` and `day` keyword arguments. See `code_template.txt` (the default
    template) as an example.
    """
    if year is None:
        year = date.today().year

    if template is None:
        template = TEMPLATE_FILE.read_text()
    else:
        template = template.read_text()

    for day in range(1, 26):
        file = Path(f"day_{day:02}.py")
        if not file.exists():
            file.write_text(template.format(year=year, day=day))


def _fix_old_inputs(year: int) -> bool:
    input_file = CONFIG_DIR / f"{year}" / "inputs.toml"

    if input_file.exists():
        inputs = tomlkit.loads(input_file.read_text())
        for day, input_ in inputs.items():
            (input_file.parent / f"{int(day):02}.txt").write_text(input_)

        input_file.unlink()


def fetch(year: int, day: int, open_with: Literal["vscode"] | None = None) -> str:
    """Fetch puzzle input.

    Inputs are cached so that an input won't be re-fetched. The `open_with` parameter
    can be used to open the fetched input as a separate file in your IDE/text editor.
    """
    _fix_old_inputs(year)

    year_dir = CONFIG_DIR / f"{year}"
    if not year_dir.exists():
        year_dir.mkdir()
    input_file = year_dir / f"{day:02}.txt"

    if input_file.exists():
        input_ = input_file.read_text()
    else:
        _wait_for_unlock(year, day)

        response = requests.get(
            url=URL.format(year=year, day=day) + "/input",
            headers=HEADERS,
            cookies=TOKEN,
        )
        if not response.ok:
            raise ValueError("Request failed.")

        input_ = response.text.rstrip()
        input_file.write_text(input_)

    if open_with is not None:
        # Open an issue to add a command to open the input file in your favorite editor.
        editors = {"vscode": ["code", "-r"]}
        subprocess.run([*editors[open_with], input_file], shell=True, check=False)

    return input_


def submit(
    year: int,
    day: int,
    part: Literal[1, 2],
    solution: Callable,
    sanity_check=True,
) -> None:
    """Submit a solution.

    Submissions are cached so that the same solution won't be resubmitted.
    """
    submissions_file = CONFIG_DIR / f"{year}" / SUBMISSIONS_FILE
    if not submissions_file.parent.exists():
        submissions_file.parent.mkdir()
    if not submissions_file.exists():
        submissions_file.touch()
    submissions = tomlkit.loads(submissions_file.read_text())
    current = submissions.setdefault(str(day), {}).setdefault(str(part), {})

    if "solution" in current:
        print(
            f"Day {day}, part {part} has already been solved. "
            f"The solution was:\n{current['solution']}"
        )
        return

    if (answer := solution()) is None:
        print("Solution produced no answer. Did you forget a return?")
        return

    answer = str(answer)

    if answer in current:
        print(
            f"Solution {answer} to part {part} has already been submitted,"
            " response was:",
        )
        _pretty_print(current[answer])
        return

    if sanity_check and input(f"Submit {answer}? [y]/n\n").startswith(("n", "N")):
        return

    while True:
        print(f"Submitting {answer} as solution to part {part}:")
        response = requests.post(
            url=URL.format(year=year, day=day) + "/answer",
            data={"level": part, "answer": answer},
            headers=HEADERS,
            cookies=TOKEN,
        )

        if not response.ok:
            raise ValueError("Request failed.")

        message = bs4.BeautifulSoup(response.text, "html.parser").article.text
        _pretty_print(message)

        if message[4] == "g":  # "You gave an answer too recently"
            minutes, seconds = re.search(r"(?:(\d+)m )?(\d+)s", message).groups()

            timeout = 60 * int(minutes or 0) + int(seconds)
            try:
                print(HIDE_CURSOR)
                while timeout > 0:
                    print(
                        f"Waiting {BOLD}{YELLOW}{timeout}{RESET} seconds to "
                        "retry...".ljust(50),
                        end="\r",
                    )
                    timeout -= 1
                    time.sleep(1)
            finally:
                print(SHOW_CURSOR)
        else:
            break

    if message[7] == "t":  # "That's the right answer! ..."
        current["solution"] = answer

        if day == 25:  # Automatically submit part 2 on day 25.
            response = requests.post(
                url=URL.format(year=year, day=day) + "/answer",
                data={"level": 2, "answer": ""},
                headers=HEADERS,
                cookies=TOKEN,
            )

        if part == 1:
            webbrowser.open(response.url)  # View part 2 in browser

    current[answer] = message
    submissions_file.write_text(tomlkit.dumps(submissions))


def _wait_for_unlock(year, day):
    """Display a countdown in the terminal to when puzzle input is available."""
    now = datetime.now().astimezone()
    unlock = datetime(year=year, day=day, **UNLOCK_TIME_INFO)

    if now < unlock:
        try:
            print(HIDE_CURSOR)
            while True:
                now = datetime.now().astimezone()
                if (delay := (unlock - now).total_seconds()) <= 0:
                    break
                print(
                    f"Waiting {BOLD}{YELLOW}{delay:.2f}{RESET} seconds for puzzle input"
                    " to unlock...".ljust(50),
                    end="\r",
                )
                time.sleep(0.1)
        finally:
            print(SHOW_CURSOR)


def _pretty_print(message):
    """Color submission response."""
    match message[7]:
        case "t":
            # "That's the right answer! ..."
            color = GREEN
        case "'" | "e":
            # "You don't seem to be solving the right level. ..."
            # "You gave an answer too recently; you have to wait ..."
            color = YELLOW
        case "n":
            # "That's not the right answer. If you're stuck, ..."
            color = RED
        case _:
            raise ValueError("Unexpected message.", message)
    print(f"{BOLD}{color}{message}{RESET}")
